#include "stm32f4xx.h"
#include "stm32f4xx_it.h"
#include "stm32f4xx_conf.h"
#include <stdio.h>
#include "usb.h"
#include "core_cm4.h"
#include "mp3dec.h"
#include "Audio.h"
#include "mp3util.h"
#include "global_status.h"
#include "io_pins.h"
#include <string.h>
#include "staticData.h"
#include "folderButtons.h"


// Macros
#define f_tell(fp)		((fp)->fptr)

//#define RESUME_IN_FOLDER

// MP3 Variables
//#define FILE_READ_BUFFER_SIZE 8192
#define FILE_READ_BUFFER_SIZE (8192*2)

struct {
	char					file_read_buffer[FILE_READ_BUFFER_SIZE];
	volatile int			bytes_left;
	char					*read_ptr;
} mp3Playback;

// Private function prototypes
static void AudioCallback(void *context,int buffer);
static void play_mp3(char* filename);
static FRESULT play_directory (uint8_t folderNr);


/*
 * Buttons
 * N1 														PE7
 * N2														PE8
 * N3														PE9
 * N4														PE10
 * N5														PE11
 * N6														PE12
 * N7														PE13
 * N8														PE14
 * N9														PE15
 * N10														PB13
 * Vol+ (timer input for encoder) 							PA0 / TIM5-ch1 --> user button - remove SB20 for lower current consumption
 * Vol- (timer input for encoder) 							PA1 / TIM5-ch2
 * Play / Pause  											PA2
 * Skip          											PA3
 * Rewind												 	PB14
 * Button of Encoder										PB15
 *
 * Power off transistor to switch off the bistable relay	PD8
 * Power Button Signal (5V switched)						PC1
 * LED Digital Protocol for RGB LEDs						PD12  (PL9823-F5, Timer4_Ch1  (Shared with onboard led1) --> PC6 change Tim3_ch1
 *
 *
 * Modifications to the board:
 * remove R26 and connect the pin which connects to C28 to the 3V backup battery
 *
 * All IO to this pins is defined in io_pins.c
 *
 * TODO: remember position in track and start from that position
 *
 */


int main(void) {
	RCC_ClocksTypeDef		RCC_Clocks;
	RCC_GetClocksFreq(&RCC_Clocks);
	SysTick_Config(RCC_Clocks.HCLK_Frequency / 1000);
	buttonsAndOutputInit();
	initStaticSram();
	globalStatusInit();
	DBGMCU_Config(DBGMCU_SLEEP, ENABLE);

	usb_init();

	for(;;) {
		usb_update();
		// TODO power off after few seconds with no usb stick

		if (usbEnumDone() >= 2) {
			usbResetEnumDone();
			globalStatusStickDetected();

			folderButtonsInit();
			while(!buttonsHandler()) {
				__WFI();
				globalStatusIdle();
			}
			while(globalStatusSelectNextDir()) {
				if( folderButtonsIsAssigned(globalStatusGetCurrentDir())) {
					play_directory(	globalStatusGetCurrentDir()	);
				}
			}
		} else {
			__WFI();
			globalStatusIdle();
		}
	}
}

static FRESULT play_directory (uint8_t folderNr) {
	char longfilename[256];
	longfilename[0]='\0';
	char shortfilename[12];
	shortfilename[0]='\0';
	bool end=false;
	bool played=false;
	FRESULT res=FR_OK;
	uint32_t currentFile=0;

	uint32_t skipFiles = staticDataGetCurrentFile(folderNr);

	do {
		if(findNextMp3(folderButtonsGetShortname(folderNr), longfilename, shortfilename)) {
			currentFile++;
			if(currentFile>=skipFiles) {
				char fullpath[255];
				sprintf(fullpath, "%s/%s", folderButtonsGetShortname(folderNr), shortfilename);
				staticDataSetPlaybackFile(folderNr,currentFile);
				play_mp3(fullpath);
				played=true;
			}
		} else {
			if(played) {
				end=true;
			}
			currentFile=0;
			skipFiles=0;
			staticDataSetPlaybackFile(folderNr,currentFile);
			longfilename[0]='\0';
			shortfilename[0]='\0';
		}
	} while ((!end) && (!globalStatusIsChangeDir()) && (globalStatusGetCurrentDir() <10));
	return res;
}


static int decodeMp3(int16_t *samples, MP3FrameInfo *mp3FrameInfo, const bool fastForward) {

	int offset, err;
	int outOfData = 0;
	offset = MP3FindSyncWord((unsigned char*)mp3Playback.read_ptr, mp3Playback.bytes_left);
	mp3Playback.bytes_left -= offset;
	mp3Playback.read_ptr += offset;

	if(fastForward) {
		err = MP3DecodeDummy(hMP3Decoder, (unsigned char**)&mp3Playback.read_ptr, (int*)&mp3Playback.bytes_left);
	} else {
		err = MP3Decode(hMP3Decoder, (unsigned char**)&mp3Playback.read_ptr, (int*)&mp3Playback.bytes_left, samples, 0);
	}
	if (err) {
		/* error occurred */
		switch (err) {
		case ERR_MP3_INDATA_UNDERFLOW:
			outOfData = 1;
			break;
		case ERR_MP3_MAINDATA_UNDERFLOW:
			/* do nothing - next call to decode will provide more mainData */
			break;
		case ERR_MP3_FREE_BITRATE_SYNC:
		default:
			outOfData = 1;
			break;
		}
	} else {
		// no error
		MP3GetLastFrameInfo(hMP3Decoder, mp3FrameInfo);

		// Duplicate data in case of mono to maintain playback speed
		if (mp3FrameInfo->nChans == 1) {
			for(int i = mp3FrameInfo->outputSamps;i >= 0;i--) 	{
				samples[2 * i]=samples[i];
				samples[2 * i + 1]=samples[i];
			}
			mp3FrameInfo->outputSamps *= 2;
		}
	}
	return(outOfData);
}

static FRESULT readPart(FIL* file, unsigned int *br, unsigned int *btr) {
	FRESULT res;

	// Copy rest of data to beginning of read buffer
	memcpy(mp3Playback.file_read_buffer, mp3Playback.read_ptr, mp3Playback.bytes_left);
	// Update read pointer for audio sampling
	mp3Playback.read_ptr = mp3Playback.file_read_buffer;

	// Read next part of file
	*btr = FILE_READ_BUFFER_SIZE - mp3Playback.bytes_left;
	res = f_read(file, mp3Playback.file_read_buffer + mp3Playback.bytes_left, *btr, br);
	staticDataSetPlaybackPosition( f_tell(file) );

	// Update the bytes left variable
	mp3Playback.bytes_left = FILE_READ_BUFFER_SIZE;
	return (res);
}

static void play_mp3(char* filename) {
	unsigned int br, btr;
	FRESULT res;
	uint32_t startPosition=0;

	globalStatusPlay();


	mp3Playback.bytes_left = FILE_READ_BUFFER_SIZE;
	mp3Playback.read_ptr = mp3Playback.file_read_buffer;

	FIL file;
	if (FR_OK == f_open(&file, filename, FA_OPEN_EXISTING | FA_READ)) {

		// Read ID3v2 Tag
		char szArtist[120];
		char szTitle[120];
		Mp3ReadId3V2Tag(&file, szArtist, sizeof(szArtist), szTitle, sizeof(szTitle));  // skip Tag no need for information
		MP3FrameInfo mp3Info = getMp3Info(&file);
		Mp3ReadId3V2Tag(&file, szArtist, sizeof(szArtist), szTitle, sizeof(szTitle));  // skip Tag no need for information

		// Fill buffer
		if((file.fsize<=staticDataGetpositionInFile()) || (staticDataGetpositionInFile()<=FILE_READ_BUFFER_SIZE)) {
			staticDataSetPlaybackPosition(0);
			startPosition=0;
		} else {
			startPosition=staticDataGetpositionInFile()-FILE_READ_BUFFER_SIZE;  // restart one element before, might be used by the following code
		}

		res=f_read(&file, mp3Playback.file_read_buffer, FILE_READ_BUFFER_SIZE, &br);
		if(res != FR_OK) {
			return;
		}
		// Play mp3
		hMP3Decoder = MP3InitDecoder();

#ifdef RESUME_IN_FOLDER
		MP3FrameInfo mp3FrameInfo;
		static int16_t dummy[4096];
		while ((f_tell(&file)<startPosition)  &&  (res == FR_OK)) {
			decodeMp3(dummy, &mp3FrameInfo, true);
			//if (mp3Playback.bytes_left < (FILE_READ_BUFFER_SIZE * 3 / 4)) { // speeds up reading a lot
			if (mp3Playback.bytes_left < 1000) { // speeds up reading a lot TODO: check with high bitrate mp3 if boundary is ok.
				res = readPart(&file, &br, &btr);
			}
		}
#endif

		InitializeAudioBySamplerate(mp3Info.samprate);

		AudioSetVolume(staticDataGetVolume());
		PlayAudioWithCallback(AudioCallback, 0);

		for(;;) {
			/*
			 * If past half of buffer, refill...
			 *
			 * When bytes_left changes, the audio callback has just been executed. This
			 * means that there should be enough time to copy the end of the buffer
			 * to the beginning and update the pointer before the next audio callback.
			 * Getting audio callbacks while the next part of the file is read from the
			 * file system should not cause problems.
			 */
			buttonsHandler();
			globalStatusUpdateVolume();

			if (globalStatusIsPlaying()) {
				if (mp3Playback.bytes_left < (FILE_READ_BUFFER_SIZE / 2)) {
					res = readPart(&file, &br, &btr);
					// Out of data or error or user button... Stop playback!
					if(globalStatusIsChangeDir()) {
						AudioSetVolume(0);
						StopAudio();

						// Re-initialize and set volume to avoid noise
						InitializeAudio(Audio44100HzSettings);
						f_close(&file);
						return;
					}

					if (br < btr || res != FR_OK || globalStatusIsSkip() ) {
						AudioSetVolume(0);
						StopAudio();

						// Re-initialize and set volume to avoid noise
						InitializeAudio(Audio44100HzSettings);
						f_close(&file);
						//						if(!globalStatusIsSkip()) {
						staticDataResetPlaybackPosition();
						//						}

						return;
					}
				}
				__WFI();
			} else {
				// power off selected
				AudioSetVolume(0);
				StopAudio();

				// Re-initialize and set volume to avoid noise
				InitializeAudio(Audio44100HzSettings);
				f_close(&file);
				return;
			}
		}
	}

}


/*
 * Called by the audio driver when it is time to provide data to
 * one of the audio buffers (while the other buffer is sent to the
 * CODEC using DMA). One mp3 frame is decoded at a time and
 * provided to the audio driver.
 */
static void AudioCallback(void *context, int buffer) {
	static int16_t audio_buffer0[4096];
	static int16_t audio_buffer1[4096];

	int outOfData = 0;
	MP3FrameInfo mp3FrameInfo;

	int16_t *samples;

	if(!globalStatusIsPlaying()) {
		return;
	}

	if (buffer) {
		samples = audio_buffer0;
	} else {
		samples = audio_buffer1;
	}

	if(globalStatusIsPause()) {
		memset(samples, 0, 4096);
		ProvideAudioBuffer(samples, 4000);
		globalStatusIdle();
	} else {
		globalStatusShowActivity();

		outOfData=decodeMp3(samples , &mp3FrameInfo, false);
		if (!outOfData) {
			ProvideAudioBuffer(samples, mp3FrameInfo.outputSamps);
		}
	}
}


